<?php

// $Id: jpeg.class.php 137 2004-07-24 14:06:59Z tea $

/**
 * jpeg.class.php 
 *
 * @package PhpJPEG
 */

/**
 * require ...
 */
require_once "jpeg.constants.php";
require_once "exif.class.php";

/**
 * @param string message
 */
function _verbose($message) {
    if (false) echo "log: $message\n";
}

/**
 * @param char 
 * @return int
 */
function readchar($data) {
    return array_shift(unpack("C", $data));
}

/**
 * @param string (2bytes BigEndian)
 * @return int
 */
function readbyte($data) {
    return array_shift(unpack("N", "\x00\x00".$data));
}

/**
 * @param string (4bytes BigEndian)
 * @return int
 */
function readlong($data) {
    return array_shift(unpack("N", $data));
}


/**
 * This class has information of JPEG.
 *
 * @author   TeaMiKL 
 * @package  PhpJPEG
 */
class JPEG_Information
{
    /**
     * @var string filename
     */
    var $filename;

    /**
     * @var int major_version
     */
    var $major_version;

    /**
     * @var int minor_version
     */
    var $minor_version;

    /**
     * @var EXIF exif
     */
    var $exif;

    /**
     * @var int width
     */
    var $width;

    /**
     * @var int height
     */
    var $height;

    /**
     * @var string comment
     */
    var $comment;

    /**
     * @var string copyright
     */
    var $copyright;

    function JPEG_Information() {
        _verbose("new JPEG_Information()");

        $this->filename = NULL;
        $this->major_version = NULL;
        $this->minor_version = NULL;
        $this->exif = NULL;
        $this->width = NULL;
        $this->height = NULL;
        $this->comment = NULL;
        $this->copyright = NULL;
    }
    
    /**
     * @return array
     *
     * @todo  Check exists EXIF field and IFD0 field.
     */
    function getEXIF() {
        return $this->exif->tiff->IFD[0];
    }
}

/**
 * JPEG Segments marker class
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG_Segment_Marker
{
    /**
     * @var Integer type
     */
    var $type;

    /**
     * @var byte data
     */
    var $data;

    /**
     * @var char first (1 byte)
     */
    var $first;

    /**
     * @var char second (1 byte)
     */
    var $second;

    /**
     * @param string marker
     * @todo JPEG_Segment_Marker is context class.
     */
    function JPEG_Segment_Marker($marker) {
        _verbose("new JPEG_Segment_Marker([marker])");
        
        # XXX
        # This code had Warning: Uncounce offset 0
        # list($dummy, $this->first, $this->second) = unpack("C2", $marker);
        $temp = unpack("C2", $marker);
        $this->first = $temp[1];
        $this->second = $temp[2];

        $this->type = $this->second;
        $this->data = $marker;
    }

    function isSOI() {
        return ($this->type == M_SOI) ? true : false;
    }
    
    function isEOI() {
        return ($this->type == M_EOI) ? true : false;
    }
    
    function isSOS() {
        return ($this->type == M_SOS) ? true : false;
    }
    
    function isDAC() {
        return ($this->type == M_DAC) ? true : false;
    }

    function isDHT() {
        return ($this->type == M_DHT) ? true : false;
    }

    function isSOF() {
        global $SOF_array;
        return in_array($this->type, $SOF_array);
    }

    function isAPP() {
        global $APP_array;
        return in_array($this->type, $APP_array);
    }

    function isCOM() {
        return ($this->type == M_COM) ? true : false;
    }
    
    function toString() {
        return array_shift(unpack("H*", $this->data));
    }
}

/**
 * This class is piece of JPEG data.
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG_Segment
{
    /**
     * @param JPEG_Segment_Marker marker
     */
    var $marker;    # 2 byte

    /**
     * @var Integer length
     */
    var $length;    # 2 byet

    /**
     * @var string text
     */
    var $text;      # length - 2 byte

    function JPEG_Segment($marker, $length=0, $text="") {
        _verbose("new JPEG_Segment");
        $this->setMarker($marker);
        $this->setLength($length);
        $this->setText($text);
    }

    function setMarker($marker) {
        $this->marker = new JPEG_Segment_Marker($marker);
    }
    
    function getMarker() {
        return $this->marker->toString();
    }
    
    function setLength($length) {
        $this->length = $length;
    }
    
    function getLength() {
        return $this->length;
    }
    
    function setText($text) {
        $this->text = $text;
    }
    
    function getText() {
        return $this->text;
    }
}


/**
 * This class read segment from raw file stream.
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG_Tokenizer
{
    var $filehandle;

    function JPEG_Tokenizer($filehandle) {
        _verbose("new JPEG_Tokenizer([filehandle])");
        
        $this->filehandle = $filehandle;
    }

    function readMarker() {
        return @fread($this->filehandle, 2);
    }

    function readLength() {
        return @fread($this->filehandle, 2);
    }

    function readText($length=0) {
        return @fread($this->filehandle, $length);
    }
}


/**
 * This class parses JPEG segments.
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG_Parser
{
    /**
     * @var JPEG_Tokenizer tokenizer
     */
    var $tokenizer;

    /**
     * @param JPEG_Tokenizer tokenizer
     */
    function JPEG_Parser($tokenizer) {
        _verbose("new JPEG_Parser");
        
        $this->tokenizer = $tokenizer;
    }

    /**
     * @return JPEG_Segment
     */
    function getNext() {
        _verbose("JPEG_Parser::getNext()");
        
        $marker = $this->tokenizer->readMarker();
        
        // TODO Exception wrong format segment.
        
        if (strlen($marker) < 2) {
            $marker = $markder->toString();
            echo "$marker is bad marker\n";
            return NULL;
        }

        $c = new JPEG_Segment_Marker($marker);

        if ($c->first != 0xFF) {
            echo "first byte is no FF\n";
            return NULL;
        }

        switch ($c->type)
        {
            case M_SOI:
            case M_EOI:
            case M_SOS:
                return new JPEG_Segment($marker);
                
            default:
                $length = readbyte($this->tokenizer->readLength());
                $text   = $this->tokenizer->readText($length - 2);
                return new JPEG_Segment($marker, $length, $text);
        }

        return NULL;
    }
}

/**
 * class JPEG_Stream
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG_Stream
{
    /**
     * @var JPEG_Parser parser
     */
    var $parser;

    /**
     * @var JPEG_Tokenizer tokenizer
     */
    var $tokenizer;

    /**
     * @var string filename
     */
    var $filename;

    /**
     * @var resource filehandle
     */
    var $filehandle;

    /**
     * @param string filename
     */
    function JPEG_Stream($filename) {
        _verbose("new JPEG_Stream()");
    
        $this->open($filename);
    }

    /**
     * @param string filename
     */
    function open($filename) {
        _verbose("JPEG_Stream::open($filename)");
        
        $this->filename   = $filename;
        $this->filehandle = @fopen($filename, "rb");
        $this->tokenizer  = new JPEG_Tokenizer($this->filehandle);
        $this->parser     = new JPEG_Parser($this->tokenizer);
    }

    /**
     * Close filehandle
     */
    function close() {
        _verbose("JPEG_Stream::close()");
        
        @fclose($this->filehandle);
    }
    
    /**
     * @return JPEG_Segment
     */
    function getNext() {
        _verbose("JPEG_Stream::getNext()");
    
        return $this->parser->getNext();
    }
}


/**
 * This is a high level JPEG manipulate class.
 *
 * @author TeaMiKL
 * @package PhpJPEG
 */
class JPEG
{
    /**
     * @var JPEG_Information info
     */
    var $info;
    
    /**
     * @var JPEG_Stream stream
     */
    var $stream;
    
    /**
     * @var string filename
     */
    var $filename;

    /**
     * @var array segments 
     */
    var $segments;

    /**
     * @param string filename
     */
    function JPEG($filename) {
        _verbose("new JPEG($filename)");
        
        $this->info = new JPEG_Information;

        if ($filename) {
            $this->filename = $filename;
            $this->process();
        }
    }
    
    /**
     * 
     */
    function process() {
        _verbose("JPEG::process()");

        global $SOF_descriptions;

        $this->segments = array();
        $this->stream = new JPEG_Stream($this->filename);
        $this->info->filename = $this->filename;
        
        while ($segment = $this->stream->getNext())
        {
            # TODO
            # $this->segments[] = $segments;
            
            $marker = $segment->marker;

            if ($marker->isSOI()) {
                /* Skip */
            }
            elseif ($marker->isEOI() || $marker->isSOS()) {
                break;
            }
            elseif ($marker->isSOF()) {
                $text = $segment->getText();
                
                $data_percision = readchar(substr($text, 0, 1));
                $image_width    = readbyte(substr($text, 1, 2));
                $image_height   = readbyte(substr($text, 3, 2));
                $component_num  = readchar(substr($text, 5, 1));

                if (array_key_exists($marker->type, $SOF_descriptions))
                    $process = $SOF_descriptions[$marker->type];
                else
                    $process = "Unknown";

                $this->info->width   = $image_width;
                $this->info->height  = $image_height;
                $this->info->process = $process;
            }
            elseif ($marker->isAPP()) {
                if ($marker->type == M_APP0) {

                }
                elseif ($marker->type == M_APP1) {
                
                    $exif = new EXIF($segment->getText());
                    
                    if ($exif->signature == "Exif") {
                        $this->info->exif = $exif;

                    }
                    else {
                        # XXX No support XML/RDF format EXIF yet.
                        # echo "No support EXIF format\n";
                    }
                    $exif = NULL;
                }
            }
            elseif ($marker->isCOM()) {
                $this->info->comment = $segment->getText();
            }
            else {

            }

            # printf( "Marker: %s\n", $marker->toString());
        }
        
        $this->stream->close();
    }
}

?>
