<?php

// $Id: tiff.class.php 137 2004-07-24 14:06:59Z tea $

/**
 * tiff.class.php
 *
 * @package PhpJPEG
 */

/**
 * Class for Image File Directory
 *
 * @author TeaMiKL
 * @package PhpJPEG
 * @todo refactoring to composite pattern.
 */
class IFD_Entry {
    /**
     * @var Integer tag
     */
    var $tag;

    /**
     * @var Integer type
     */
    var $type;

    /**
     * @var Integer number (or length)
     */
    var $number;
    
    /**
     * @var Integer value (or offset)
     */
    var $value;

    /**
     * @param Integer tag
     * @param Integer type
     * @param Integer number or length
     * @param Integer value or offset
     */
    function IFD_Entry($tag, $type, $number, $value=NULL) {
        $this->tag    = $tag;
        $this->type   = $type;
        $this->number = $number;
        $this->value  = $value;
    }
}


/**
 * TIFF Stream class
 *
 * @author TeaMiKL
 * @package PhpJPEG
 *
 * @todo refactoring design of classes.
 */
class TIFF_Stream {

    /**
     * @var string data
     */
    var $data;

    /**
     * @var string endian
     */
    var $endian;

    /**
     * @var Integer offset
     */
    var $offset;

    /**
     * @var Integer padding
     */
    var $padding;

    /**
     * @param String data
     */
    function TIFF_Stream($data) {
        $this->data    = $data;
        $this->offset  = 2; # increment for the length of endian.
        $this->endian  = $this->read(0, 2);
        $this->padding = $this->readbyte();
    }

    /**
     *
     */
    function close() {
        $this->data = NULL;
    }

    /*
     * @return String
     */
    function toString() {
        return sprintf("Offset: %d, Endian: %s\n", $this->offset, $this->endian);
    }

    /**
     * @param string endian
     */
    function setEndian($endian) {
        $this->endian = $endian;
    }

    /**
     * @return string endian
     */
    function getEndian() {
        return $this->endian;
    }

    /**
     * @param Integer offset
     */
    function setOffset($offset) {
        $this->offset = $offset;
    }

    /**
     * @return Integer offset
     */
    function getOffset() {
        return $this->offset;
    }

    /**
     * @return boolean
     */
    function isIntelByteOrder() { # "II"
        return ($this->endian == "II") ? true : false;
    }

    /**
     * @return boolean
     */
    function isMotorolaByteOrder() { # "MM"
        return ($this->endian == "MM") ? true : false;
    }

    /**
     * @param Integer offset
     * @param Integer offset
     * @return String 
     */
    function read($offset, $length) {
        return substr($this->data, $offset, $length);
    }

    /**
     * @return Integer
     */
    function readbyte() {
        $temp = substr($this->data, $this->offset, 2);
        $this->offset += 2;

        if ($this->isIntelByteOrder())
            return array_shift(unpack("V", $temp."\x00\x00"));
            
        if ($this->isMotorolaByteOrder())
            return array_shift(unpack("N", "\x00\x00".$temp));
    
        return NULL;
    }

    /**
     * @return Integer
     */
    function readlong() {
        $temp = substr($this->data, $this->offset, 4);
        $this->offset += 4;
    
        if ($this->isIntelByteOrder())
            return array_shift(unpack("V", $temp));
            
        if ($this->isMotorolaByteOrder())
            return array_shift(unpack("N", $temp));
    
        return NULL;
    }

    /**
     * @return boolean
     * @todo implement
     */
    function hasNextDirectory() {
        /* XXX */ die("No implemented, yet.");
    }

    /**
     * @return boolean
     * @todo implement
     */
    function hasNextEntry() {
        /* XXX */ die("No implemented, yet.");
    }

    /**
     * @return IFD_Entry
     * @todo support types
     * @todo fix bug. you must get values $number times.
     */
    # XXX Set offset first.
    function getNextEntry() {
        $tag           = $this->readbyte();
        $type          = $this->readbyte();
        $length_number = $this->readlong();
        $value_offset  = $this->readlong();
        $entry         = new IFD_Entry($tag, $type, $length_number);
        
        switch ($type) {
        case 1: # unsigned byte
            $entry->value = $value_offset;
            break;

        case 2: # ASCII strings
            $entry->value = $this->read($value_offset, $length_number-1);
            break;

        case 3: # unsigned short
        case 4: # unsigned long
            $entry->value = $value_offset;
            break;

        # TODO
        case 5: # Rational
        case 6:
        case 7:
        case 8:
        case 9:
        case 10;
        case 11:
        case 12:
            break;

        default:
            # XXX Invalid type of IFD.
            die("Invalid number of IFD type: $type\n");
            break;
        }

        return $entry;
    }
}

/**
 * Class for parse TIFF data.
 *
 * @author
 * @package PhpJPEG
 */
class TIFF {

    /**
     * @var array IFD (Array of Image File Directory)
     */
    var $IFD;
    
    /**
     * @param String data
     */
    function TIFF($data) {
        $this->IFD = array();
        $this->stream = new TIFF_Stream($data);
        $this->parse_directories();
        $this->stream->close();
    }

    function parse_directories() {
        while (($offset = $this->stream->readlong()) > 0)
            array_push($this->IFD, $this->parse_directory_entries($offset)); 
    }

    /**
     * @param int offset
     * @param string infotype (exif or gps)
     */
    function parse_directory_entries($offset, $infotype="exif") {
        global $EXIF_INFO, $GPS_INFO;

        # XXX set information array
        $info = ($infotype == "gps") ? $GPS_INFO : $EXIF_INFO;

        $this->stream->setOffset($offset);
        $number = $this->stream->readbyte();
        $entries = array();
    
        for ($i = 1; $i <= $number; ++$i) {
            $entry = $this->stream->getNextEntry();
            
            if (0) { # XXX DEBUG
                printf("[%2d] - %4x %4d %4d %s\n",
                    $i, $entry->tag, $entry->type, $entry->number, $entry->value);
                if (array_key_exists($entry->tag, $EXIF_INFO)) 
                    printf("%s\n", $EXIF_INFO[$entry->tag]);
            }

            if (array_key_exists($entry->tag, $info)) {
                $entries[$info[$entry->tag]] = $entry->value;

                # XXX BAD Design
                if ($infotype == "exif") {
                    if ($entry->tag == EXIF_TAG_EXIF_IFD_POINTER ||
                        $entry->tag == EXIF_TAG_GPS_INFO_IFD_POINTER) {
                        $save_offset = $this->stream->getOffset();
                        $element = $this->parse_directory_entries(
                                   $entry->value,
                                   ($entry->tag == EXIF_TAG_GPS_INFO_IFD_POINTER)
                                    ? "gps"
                                    : "exif");
                        $entries[$info[$entry->tag]] = $element;
                        $this->stream->setOffset($save_offset);
                        $save_offset = 0;
                        $element = NULL;
                    }
                }
            }
            else {
                if (0) # XXX DEBUG
                    printf("Unknown tag: %x\n", $entry->tag);
            }

        }

        return $entries;
    }
}

?>
